package com.annimon.ownlang.parser.ast;

import com.annimon.ownlang.exceptions.OperationIsNotSupportedException;
import com.annimon.ownlang.lib.NumberValue;
import com.annimon.ownlang.lib.StringValue;
import com.annimon.ownlang.lib.Types;
import com.annimon.ownlang.lib.Value;
import com.annimon.stream.Objects;

/**
 *
 * @author aNNiMON
 */
public final class UnaryExpression implements Expression, Statement {

    public static final int
        INCREMENT_PREFIX = 1,//("++"),
        DECREMENT_PREFIX = 2,//("--"),
        INCREMENT_POSTFIX = 3,//("++"),
        DECREMENT_POSTFIX = 4,//("--"),
        NEGATE = 5,//("-"),
        // Boolean
        NOT = 6,//("!"),
        // Bitwise
        COMPLEMENT = 7,//("~");
        LAST = 7;
    
    private static final String[] OPERATION_NAMES = {
        "", "++", "--", "++", "--", "-", "!", "~"
    };
    
    public static String operationName(int operation) {
        if (operation < 1 && operation > LAST) return "";
        return OPERATION_NAMES[operation];
    }
    
    public final Expression expr1;
    public final int operation;

    public UnaryExpression(int operation, Expression expr1) {
        this.operation = operation;
        this.expr1 = expr1;
    }
    
    /*@Override*/
    public void execute() {
        eval();
    }
    
    /*@Override*/
    public Value eval() {
        final Value value = expr1.eval();
        switch (operation) {
            case INCREMENT_PREFIX: {
                if (expr1 instanceof Accessible) {
                    return ((Accessible) expr1).set(increment(value));
                }
                return increment(value);
            }
            case DECREMENT_PREFIX: {
                if (expr1 instanceof Accessible) {
                    return ((Accessible) expr1).set(decrement(value));
                }
                return decrement(value);
            }
            case INCREMENT_POSTFIX: {
                if (expr1 instanceof Accessible) {
                    ((Accessible) expr1).set(increment(value));
                    return value;
                }
                return increment(value);
            }
            case DECREMENT_POSTFIX: {
                if (expr1 instanceof Accessible) {
                    ((Accessible) expr1).set(decrement(value));
                    return value;
                }
                return decrement(value);
            }
            case NEGATE: return negate(value);
            case COMPLEMENT: return complement(value);
            case NOT: return not(value);
            default:
                throw new OperationIsNotSupportedException(operationName(operation));
        }
    }
    
    private Value increment(Value value) {
        if (value.type() == Types.NUMBER) {
            final Object/*Number*/ number = (Object/*Number*/) value.raw();
            if (number instanceof Double) {
                return new NumberValue(Objects.doubleValue(number) + 1);
            }
            if (number instanceof Float) {
                return new NumberValue(Objects.floatValue(number) + 1);
            }
            if (number instanceof Long) {
                return new NumberValue(Objects.longValue(number) + 1);
            }
        }
        return new NumberValue(value.asInt() + 1);
    }
    
    private Value decrement(Value value) {
        if (value.type() == Types.NUMBER) {
            final Object/*Number*/ number = (Object/*Number*/) value.raw();
            if (number instanceof Double) {
                return new NumberValue(Objects.doubleValue(number) - 1);
            }
            if (number instanceof Float) {
                return new NumberValue(Objects.floatValue(number) - 1);
            }
            if (number instanceof Long) {
                return new NumberValue(Objects.longValue(number) - 1);
            }
        }
        return new NumberValue(value.asInt() - 1);
    }
    
    private Value negate(Value value) {
        if (value.type() == Types.STRING) {
            final StringBuffer sb = new StringBuffer(value.asString());
            return new StringValue(sb.reverse().toString());
        }
        if (value.type() == Types.NUMBER) {
            final Object/*Number*/ number = (Object/*Number*/) value.raw();
            if (number instanceof Double) {
                return new NumberValue(-Objects.doubleValue(number));
            }
            if (number instanceof Float) {
                return new NumberValue(-Objects.floatValue(number));
            }
            if (number instanceof Long) {
                return new NumberValue(-Objects.longValue(number));
            }
        }
        return new NumberValue(-value.asInt());
    }
    
    private Value complement(Value value) {
        if (value.type() == Types.NUMBER) {
            final Object/*Number*/ number = (Object/*Number*/) value.raw();
            if (number instanceof Long) {
                return new NumberValue(~Objects.longValue(number));
            }
        }
        return new NumberValue(~value.asInt());
    }
    
    private Value not(Value value) {
        return NumberValue.fromBoolean(value.asInt() == 0);
    }
    
    /*@Override*/
    public void accept(Visitor visitor) {
        visitor.visit(this);
    }
    
    /*@Override*/
    public String toString() {
        switch (operation) {
            case INCREMENT_POSTFIX:
            case DECREMENT_POSTFIX:
                return expr1 + " " + operationName(operation);
            default:
                return operationName(operation) + " " + expr1;
        }
    }
}
